/* -*- linux-c -*- */


/*
    lm75.c - Part of a Linux module for reading sensor data.
    Copyright (c) 1998  Philip Edelbrock <phil@netroedge.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*
 *  LM75 sensor driver for LINUX
 *
 * Copyright (c) 1998 Philip Edelbrock
 * 
 * part of a module which provides a 
 * /proc/sensors file which shows
 * some information about your system.
 * Voltages, temperature and fan speeds
 * are currently reported.
 *
 *
 * please report any bugs/patches to me:
 *  phil@netroedge.com
 *
 *
 * Reference:
 *   National Semiconductor
 *   LM75 Digital Temperature Sensor and Thermal Watchdog with
 *   Two-Wire Interface, PDF, April 1997.
 *   http://www.national.com/ds/LM/LM75.pdf
 *
 *   Intel
 *   82371AB PCI-TO-ISA / IDE XCELERATOR (PIIX4), PDF, April 1997.
 *   ftp://ftp.intel.com
 *   
 *   System Management Bus Specification, Rev. 1.0, 1996.
 *   Contributors:
 *   Benchmarq Microelectronics Inc ., Duracell Inc.,
 *   Energizer Power Systems, Intel Corporation, Linear Technology Corporation,
 *   Maxim Integrated Products, Mitsubishi Electric Corporation,
 *   National Semiconductor Corporation, Toshiba Battery Co.,
 *   Varta Batterie AG, All rights reserved.
 *   http://www.sbs-forum.org/
 */

#include <linux/config.h>

/* The next define disables the redefinition of kernel_version. As it is
   already defined and initialized in lm78.c, this would result in twice
   defined symbols. We must include <linux/version.h> by hand now. */
#define __NO_VERSION__

/* Hack to allow the use of CONFIG_MODVERSIONS. In the kernel itself, this
   is arranged from the main Makefile. */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/bios32.h>
#include <asm/io.h>
#include <linux/pci.h>
#include "lm75.h"

/* These are functions used externally */
extern int Init_LM75(void);
extern void Cleanup_LM75(void);
extern int print_lm75_values (char *buf);


/* These are functions used only internally (but could be
   used externally in the future)*/
static int read_smbus_word(int addr, int command);
static int write_smbus_byte(int addr, long value);
static int write_smbus_word(int addr, long value);
/* Time killer */
static void our_usleep(int wait_time);

/* Global variable */
unsigned short PIIX4_smba = 0;


/* This kill some time in a baaaaaad way. :') Needs to be replaced! */
static void our_usleep(int wait_time) {
long killtime;
long temp;
 for (; wait_time > 0; wait_time--) {
         for (killtime=0; killtime<=4000; killtime++) {
                temp=0xAABC; temp= (temp << 13);
         }
  }
}

/*
 *      Initialize SMBus code.
 */
int Init_LM75(void){
int error_return=0;
unsigned char  PIIX4_bus, PIIX4_devfn;

    if (pcibios_present() == 0) {
        printk("LM78: SMBus: Error: No PCI-bus found!\n");
        error_return=1;
    } else {    
        if(pcibios_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371AB_0, 0,
                               &PIIX4_bus, &PIIX4_devfn)) {
            printk("LM78: Error: Can't detect PIIX4!\n");
            error_return=1;
        } else {
            pcibios_read_config_word(PIIX4_bus, PIIX4_devfn | 3, 0x90, &PIIX4_smba);
            PIIX4_smba &= 0xfff0;   
            if (check_region(PIIX4_smba, 7)) {
	    	printk("LM78: SMBus: Error: Use PIIX4_smb (already in use!\n");
                error_return=1; 
	    } else
            	request_region(PIIX4_smba, 7, "lm78");
        }   
    }   
    if (error_return)
        printk("LM78: SMBus setup failed.  LM75 readings will be disabled.\n");
#ifdef DEBUG
    else
    	printk("PIIX4_smba = 0x%X\n",PIIX4_smba);
#endif
    return error_return;
}

/*
 *      Write byte value to SMBus device at given address.
 */
int write_smbus_byte(int addr, long value) {

int HOST_BUSY;
int timeout=0;
int result=0;

/* Setup protocol type, address, device command, data, and control bits */
    outb_p( value & 0x0FF, SMBHSTCMD);
    outb_p( (value & 0x0FF00) >> 8, SMBHSTDAT0);
    outb_p( (value & 0x0FF0000) >> 16, SMBHSTDAT1); /* Not nessesary, but doesn't hurt */
    outb_p( 0x08, SMBHSTCNT); /* SMBus operation size = Byte */
    if (inb_p(SMBHSTSTS) != 0x00) {
#ifdef DEBUG
	printk("LM78: write_smbus_byte: SMBus busy. Resetting... ");
#endif
        outb_p( inb(SMBHSTSTS), SMBHSTSTS);
#ifdef DEBUG
	if (inb_p(SMBHSTSTS) != 0x00)
		printk("Failed!\n");
	else
		printk("Successfull!\n");
#endif
    }
    outb_p( (addr << 1) & 0x0FE, SMBHSTADD);
/* address of possible thermal sensors are from
 1001000x to 1001111x ( x = r/w bit ) */

/* Set the start bit */
    outb_p(inb(SMBHSTCNT) | 0x040, SMBHSTCNT); 
/* start the transaction by setting bit 6 */

/* Wait for a fraction of a second! */
   our_usleep(10);

/* Poll Host_busy bit */
   HOST_BUSY=inb_p(SMBHSTSTS) & 0x01;
   while (HOST_BUSY & (timeout++ < MAX_TIMEOUT)) {
/* Wait for a while and try again*/
        our_usleep(5);
	HOST_BUSY=inb_p(SMBHSTSTS) & 0x01;
   }
#ifdef DEBUG
   if (timeout++ >= MAX_TIMEOUT)
	   printk("LM78: write_smbus_byte: SMBus Timeout!\n");
   printk("LM78: write_smbus_byte: time cycles for transaction: %d (0x%X)\n",timeout,addr);
#endif

/* Sense for no device at the address */
   if (inb_p(SMBHSTSTS) == 0x04)  {
        result=-1;
#ifdef DEBUG
	printk("LM78: write_smbus_byte: SMBus error: no response from 0x%X !\n",addr);
#endif
   }
#ifdef DEBUG
   if (inb_p(SMBHSTSTS) & 0x010)
	printk("LM78: write_smbus_byte: SMBus error: Failed bus transaction (0x%X).\n",addr);
   if (inb_p(SMBHSTSTS) & 0x08)
	printk("LM78: write_smbus_byte: SMBus error: Bus collision (0x%X).\n",addr);
   if (inb_p(SMBHSTSTS) & 0x04)
	printk("LM78: write_smbus_byte: SMBus error: Device Error (0x%X).\n",addr);
#endif
    if (inb_p(SMBHSTSTS) != 0x00)
        outb_p( inb(SMBHSTSTS), SMBHSTSTS);
   return result;
}

/*
 *      Write byte value to SMBus device at given address.
 */
int write_smbus_word(int addr, long value) {

int HOST_BUSY;
int timeout=0;
int result=0;

/* Setup protocol type, address, device command, data, and control bits */
    outb_p( value & 0x0FF, SMBHSTCMD);
    outb_p( (value & 0x0FF00) >> 8, SMBHSTDAT0);
    outb_p( (value & 0x0FF0000) >> 16, SMBHSTDAT1);
    outb_p( 0x0C, SMBHSTCNT); /* SMBus operation size = Byte */
    if (inb_p(SMBHSTSTS) != 0x00) {
#ifdef DEBUG
	printk("LM78: write_smbus_byte: SMBus busy. Resetting... ");
#endif
        outb_p( inb(SMBHSTSTS), SMBHSTSTS);
#ifdef DEBUG
	if (inb_p(SMBHSTSTS) != 0x00)
		printk("Failed!\n");
	else
		printk("Successfull!\n");
#endif
    }
    outb_p( (addr << 1) & 0x0FE, SMBHSTADD);
/* address of possible thermal sensors are from
 1001000x to 1001111x ( x = r/w bit ) */

/* Set the start bit */
    outb_p(inb(SMBHSTCNT) | 0x040, SMBHSTCNT); 
/* start the transaction by setting bit 6 */

/* Wait for a fraction of a second! */
   our_usleep(10);

/* Poll Host_busy bit */
   HOST_BUSY=inb_p(SMBHSTSTS) & 0x01;
   while (HOST_BUSY & (timeout++ < MAX_TIMEOUT)) {
/* Wait for a while and try again*/
        our_usleep(5);
	HOST_BUSY=inb_p(SMBHSTSTS) & 0x01;
   }
#ifdef DEBUG
   if (timeout++ >= MAX_TIMEOUT)
	   printk("LM78: write_smbus_byte: SMBus Timeout!\n");
   printk("LM78: write_smbus_byte: time cycles for transaction: %d (0x%X)\n",timeout,addr);
#endif

/* Sense for no device at the address */
   if (inb_p(SMBHSTSTS) == 0x04)  {
        result=-1;
#ifdef DEBUG
	printk("LM78: write_smbus_byte: SMBus error: no response from 0x%X !\n",addr);
#endif
   }
#ifdef DEBUG
   if (inb_p(SMBHSTSTS) & 0x010)
	printk("LM78: write_smbus_byte: SMBus error: Failed bus transaction (0x%X).\n",addr);
   if (inb_p(SMBHSTSTS) & 0x08)
	printk("LM78: write_smbus_byte: SMBus error: Bus collision (0x%X).\n",addr);
   if (inb_p(SMBHSTSTS) & 0x04)
	printk("LM78: write_smbus_byte: SMBus error: Device Error (0x%X).\n",addr);
#endif
    if (inb_p(SMBHSTSTS) != 0x00)
        outb_p( inb(SMBHSTSTS), SMBHSTSTS);
   return result;
}


/*
 *      Read value from SMBus device at given address.
 */
int read_smbus_word(int addr, int command) {

int HOST_BUSY;
int timeout=0;
int result=0;

/* Setup protocol type, address, device command, data, and control bits */
    outb_p( command & 0x0FF, SMBHSTCMD);
    outb_p( 0x00, SMBHSTDAT0);
    outb_p( 0x00, SMBHSTDAT1);
    outb_p( 0x0C, SMBHSTCNT); /* SMBus operation size = Word */
    if (inb_p(SMBHSTSTS) != 0x00) {
        outb_p( inb(SMBHSTSTS), SMBHSTSTS);
    }
    outb_p( (addr << 1) | 0x01, SMBHSTADD);
/* address of possible thermal sensors are from
 1001000x to 1001111x ( x = r/w bit ) */

/* Set the start bit */
    outb_p(inb(SMBHSTCNT) | 0x040, SMBHSTCNT); 
/* start the transaction by setting bit 6 */

/* Wait for a fraction of a second! */
   our_usleep(10);

/* Poll Host_busy bit */
   HOST_BUSY=inb_p(SMBHSTSTS) & 0x01;
   while (HOST_BUSY & (timeout++ < MAX_TIMEOUT)) {
/* Wait for a while and try again*/
        our_usleep(5);
	HOST_BUSY=inb_p(SMBHSTSTS) & 0x01;
   }
   
#ifdef DEBUG
   if (timeout++ >= MAX_TIMEOUT)
	   printk("LM78: read_smbus_word: SMBus Timeout!\n");
   printk("LM78: read_smbus_word: time cycles for transaction: %d (0x%X)\n",timeout,addr);
#endif

   if (! (inb(SMBHSTSTS) & 0x01)) {
        result=(inb(SMBHSTDAT0) << 8) | inb(SMBHSTDAT1);
   }

/* Sense for no device at the address */
   if (inb_p(SMBHSTSTS) == 0x04)  {
        result=-1;
#ifdef DEBUG
	printk("LM78: read_smbus_word: SMBus error: no response from 0x%X !\n",addr);
#endif
   }
#ifdef DEBUG
   if (inb_p(SMBHSTSTS) & 0x010)
	printk("LM78: read_smbus_word: SMBus error: Failed bus transaction (0x%X).\n",addr);
   if (inb_p(SMBHSTSTS) & 0x08)
	printk("LM78: read_smbus_word: SMBus error: Bus collision (0x%X).\n",addr);
   if (inb_p(SMBHSTSTS) & 0x04)
	printk("LM78: read_smbus_word: SMBus error: Device Error (0x%X).\n",addr);
#endif
   if (inb_p(SMBHSTSTS) != 0x00)
        outb_p( inb(SMBHSTSTS), SMBHSTSTS);
   return result;
}



int print_lm75_values (char *buf) {

int i;
int len=0;
int value;

    /* Probe each possible valid LM75 address and report the temp for each found */
       for (i=0x048; i<=0x04F; i++) { 
/*	for (i=0x04C; i<=0x04C; i++) {  */
		write_smbus_byte(i,0x000001); /* Clear configuration reg */
/*		write_smbus_word(i,0x000002);
		write_smbus_word(i,0x000003); */
	        value=read_smbus_word(i,0);
	        if (value != -1) {
	           len = sprintf(buf+len, "LM75 address 0x%X reports %d.%d degrees C\n",i,
			value  >> 8, ((value >> 7) & 1) * 5);
	        }
#ifdef DEBUG
	        value=read_smbus_word(i,1);
	        if (value != -1) {
	           len += sprintf(buf+len, "LM75 reg 1 (configuration register) at address 0x%X reports 0x%X\n",i,
			value);
	        }
	        value=read_smbus_word(i,2);
	        if (value != -1) {
	           len += sprintf(buf+len, "LM75 reg 2 (Thyst) at address 0x%X reports %d.%d degrees C (0x%X).\n",i,
			value >> 8, ((value >> 7) & 1) * 5, value);
	        }
	        value=read_smbus_word(i,3);
	        if (value != -1) {
	           len += sprintf(buf+len, "LM75 reg 3 (Tos) at address 0x%X reports %d.%d degrees C (0x%X).\n",i,
			value >> 8, ((value >> 7) & 1) * 5,value);
	        }
#endif
	}
	return len;
}

/*
 *      Clean-up the LM75/SMBus code.
 */
void Cleanup_LM75(void){
        release_region(PIIX4_smba, 7); 
}
